import "core-js/modules/es6.function.name";
import "core-js/modules/es6.regexp.search";
import "core-js/modules/es6.object.assign";
import _toConsumableArray from "@babel/runtime/helpers/toConsumableArray";

/**
 * @license
 * Copyright © 2017-2018 Moov Corporation.  All rights reserved.
 */
var _targets = []; // When false, any fired events will be put in a queue until activate() is called

var activated = false; // events accumulate here when activated is false

var queue = [];
/**
 * Configures the PWA to broadcast events to all specified targets.  Once configured, You can call any method on
 * on all targets by calling `analytics.fire(method, ...params)`.
 *
 * Example:
 *
 * ```js
 *  // src/analytics.js
 *
 *  import { configureAnalytics } from 'react-storefront/analytics'
 *  import GoogleTagManagerTarget from 'react-storefront-analytics/GoogleTagManagerTarget'
 *  import MerkleSearchTarget from 'react-storefront-analytics/MerkleSearchTarget'
 *
 *  configureAnalytics(
 *    new GoogleTagManagerTarget({ apiKey: 'my_gtm_api_key' }),
 *    new MerkleSearchTarget({ apiKey: 'my_merkle_api_key' })
 *  )
 * ```
 *
 * Then, to broadcast an analytics event to all targets:
 *
 * ```js
 * import analytics from 'react-storefront/analytics'
 *
 * // this will call the 'someEvent(data)' method on all configured targets and pass { foo: 'bar' } as the data argument.
 * analytics.fire('someEvent', { foo: 'bar' })
 * ```
 * @private
 * @param {AnalyticsTarget[]} targets An array of targets to notify when analytics events occur
 * @return {AnalyticsProvider}
 */

export function configureAnalytics() {
  for (var _len = arguments.length, targets = new Array(_len), _key = 0; _key < _len; _key++) {
    targets[_key] = arguments[_key];
  }

  _targets = targets;
}
/**
 * @private
 */

export function getTargets() {
  return _targets;
}
/**
 * Stops queuing and immediatley fires all queued events
 * @private
 */

export function activate() {
  activated = true;

  for (var _i = 0; _i < queue.length; _i++) {
    var call = queue[_i];
    fire.apply(void 0, [call.event].concat(_toConsumableArray(call.args)));

    if (process.env.NODE_ENV === 'development') {
      var _console;

      (_console = console).log.apply(_console, ['[analytics]', 'sending queued event', call.event].concat(_toConsumableArray(call.args)));
    }
  }
}

function fire(event, data) {
  for (var _len2 = arguments.length, args = new Array(_len2 > 2 ? _len2 - 2 : 0), _key2 = 2; _key2 < _len2; _key2++) {
    args[_key2 - 2] = arguments[_key2];
  }

  data = data || {};

  if (typeof document !== 'undefined' && !data.metadata) {
    Object.assign(data, {
      metadata: {
        title: document.title,
        pathname: location.pathname,
        search: location.search,
        uri: location.pathname + location.search,
        referrer: document.referrer
      }
    });
  }

  args = [data].concat(_toConsumableArray(args));

  if (activated) {
    for (var _i2 = 0; _i2 < _targets.length; _i2++) {
      var target = _targets[_i2];
      var fn = target[event];

      if (fn) {
        try {
          fn.apply(target, args);
        } catch (e) {
          console.warn('Error thrown by analytics target, event=', event, 'target=', target, 'args=', args, e);
        }
      } else {
        if (typeof event === 'string') {
          console.warn("".concat(target.constructor.name, " does not support ").concat(event));
        }
      }
    }
  } else {
    queue.push({
      event: event,
      args: args
    });
  }
}
/**
 * This module's default export is an `EventEmitter` for broadcasting analytics events to all targets configured via `<AnalyticsProvider>`.
 *
 * Example:
 *
 * ```js
 * import analytics from 'react-storefront/analytics'
 *
 * // this will call the 'someEvent(data)' method on all configured targets and pass { foo: 'bar' } as the data argument.
 * analytics.fire('someEvent', { foo: 'bar' })
 * ```
 *
 * @type {EventEmitter}
 */


var analytics = {
  fire: fire
};

if (typeof Proxy !== 'undefined') {
  analytics = new Proxy({
    fire: fire
  }, {
    get: function get(o, method) {
      if (method === 'fire') {
        return fire;
      } else if (method === 'toString') {
        return function () {
          return 'AnalyticsProxy';
        };
      } else {
        return fire.bind(null, method);
      }
    }
  });
}

export default analytics;
//# sourceMappingURL=analytics.js.map